import random
from unittest import TestCase
import asyncio

from hypothesis import given
import hypothesis.strategies as st

from .mock_world import MockWorld

from simulation.location import Location
from simulation.pickups import (
    HealthPickup, AVATAR_HEALTH_MAX, HEALTH_RESTORE_DEFAULT, HEALTH_RESTORE_MAX
)


class TestHealthPickupAndEffects(TestCase):
    def setUp(self):
        """
        Mock a game for each test individually. MockWorld() will set up a:
        avatar manager, game state, turn manager and a map generator.
        """
        self.game = MockWorld()
        self.game.game_state.add_avatar(1, Location(0, 0))
        self.cell = self.game.game_state.world_map.get_cell(Location(1, 0))
        self.initial_health = self.game.avatar_manager.get_avatar(1).health
        self.loop = asyncio.get_event_loop()

    def test_health_pickups_and_effects_apply_default(self):
        """
        HealthPickups without any parameter provided.
        """
        self.cell.pickup = HealthPickup(self.cell)

        self.loop.run_until_complete(self.game.simulation_runner.run_single_turn(self.game.avatar_manager.get_player_id_to_serialised_action()))

        self.assertEqual(self.cell.avatar, self.game.avatar_manager.get_avatar(1))
        self.assertEqual(self.cell.avatar.health, self.initial_health +
                         HEALTH_RESTORE_DEFAULT)

    @given(st.integers(1, HEALTH_RESTORE_MAX))
    def test_health_pickups_and_effects_apply_custom_integers(self, restore_value):
        """
        HealthPickups with explicit integer parameter provided.
        """
        self.setUp()
        self.cell.pickup = HealthPickup(self.cell, restore_value)

        self.loop.run_until_complete(self.game.simulation_runner.run_single_turn(self.game.avatar_manager.get_player_id_to_serialised_action()))
        self.assertEqual(self.cell.avatar, self.game.avatar_manager.get_avatar(1))

        if self.initial_health + restore_value > HEALTH_RESTORE_MAX:
            expected_result_health = HEALTH_RESTORE_MAX
        else:
            expected_result_health = self.initial_health + restore_value

        self.assertEqual(self.cell.avatar.health, expected_result_health)

    @given(st.floats(1, HEALTH_RESTORE_MAX))
    def test_health_pickups_and_effects_apply_custom_floats(self, restore_value):
        """
        HealthPickups with explicit float parameter provided.
        """
        self.setUp()
        self.cell.pickup = HealthPickup(self.cell, restore_value)

        self.loop.run_until_complete(self.game.simulation_runner.run_single_turn(self.game.avatar_manager.get_player_id_to_serialised_action()))
        self.assertEqual(self.cell.avatar, self.game.avatar_manager.get_avatar(1))

        if self.initial_health + restore_value > HEALTH_RESTORE_MAX:
            expected_result_health = HEALTH_RESTORE_MAX
        else:
            expected_result_health = self.initial_health + int(round(restore_value))

        self.assertEqual(self.cell.avatar.health, expected_result_health)

    @given(st.integers(96, HEALTH_RESTORE_MAX))
    def test_health_effect_is_capped_at_HEALTH_RESTORE_MAX(self, restore_value):
        """
        Make sure health cannot go above the maximum cap. Avatar health begins at 5hp,
        so every pickup in the 96-HEALTH_RESTORE_MAX range would cause it to go above
        HEALTH_RESTORE_MAX, until the initial health changes.
        """
        self.setUp()
        self.cell.pickup = HealthPickup(self.cell, restore_value)

        self.loop.run_until_complete(self.game.simulation_runner.run_single_turn(self.game.avatar_manager.get_player_id_to_serialised_action()))

        self.assertEqual(self.cell.avatar, self.game.avatar_manager.get_avatar(1))
        self.assertEqual(self.cell.avatar.health, AVATAR_HEALTH_MAX)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('aimmo', '0004_auto_20160808_1511'),
    ]

    operations = [
        migrations.CreateModel(
            name='LevelAttempt',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('level_number', models.IntegerField()),
            ],
        ),
        migrations.AddField(
            model_name='game',
            name='completed',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='game',
            name='generator',
            field=models.CharField(default=b'Main', max_length=20, choices=[(b'Main', b'Open World'), (b'Level1', b'Level 1')]),
        ),
        migrations.AddField(
            model_name='game',
            name='main_user',
            field=models.ForeignKey(related_name='games_for_user', blank=True, to=settings.AUTH_USER_MODEL, null=True),
        ),
        migrations.AlterField(
            model_name='game',
            name='owner',
            field=models.ForeignKey(related_name='owned_games', blank=True, to=settings.AUTH_USER_MODEL, null=True),
        ),
        migrations.AddField(
            model_name='levelattempt',
            name='game',
            field=models.OneToOneField(to='aimmo.Game'),
        ),
        migrations.AddField(
            model_name='levelattempt',
            name='user',
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterUniqueTogether(
            name='levelattempt',
            unique_together=set([('level_number', 'user')]),
        ),
    ]



import requests


class DjangoCommunicator(object):
    """
    This class encapsulates the communication between aimmo-game
    and the django server
    """
    def __init__(self, django_api_url, completion_url):
        self.django_api_url = django_api_url
        self.completion_url = completion_url

    def get_game_metadata(self):
        return requests.get(self.django_api_url).json()

    def mark_game_complete(self, data=None):
        return requests.post(requests.post(self.completion_url, json=data))



import ast
import json

from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.test import Client, TestCase

from aimmo import models, app_settings

app_settings.GAME_SERVER_URL_FUNCTION = lambda game_id: ('base %s' % game_id, 'path %s' % game_id)
app_settings.GAME_SERVER_PORT_FUNCTION = lambda game_id: 0
app_settings.GAME_SERVER_SSL_FLAG = True


class TestViews(TestCase):
    CODE = 'class Avatar: pass'

    EXPECTED_GAMES = {
        'main': {
            'parameters': [],
            'main_avatar': 1,
            'users': [
                {
                    'id': 1,
                    'code': CODE,
                },
                {
                    'id': 2,
                    'code': 'test2',
                },
                {
                    'id': 3,
                    'code': 'test3',
                },
            ]
        }
    }

    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create_user('test', 'test@example.com', 'password')
        cls.user.is_staff = True
        cls.user.save()
        cls.game = models.Game(id=1, name='test', public=True)
        cls.game.save()

    def setUp(self):
        self.game.refresh_from_db()

    def login(self):
        c = Client()
        c.login(username='test', password='password')
        return c

    def _go_to_page(self, link, kwarg_name, game_id):
        c = Client()
        response = c.get(reverse(link, kwargs={kwarg_name: game_id}))
        return response

    def _make_game_private(self):
        self.game.public = False
        self.game.save()

    def test_add_new_code(self):
        c = self.login()
        response = c.post(reverse('aimmo/code', kwargs={'id': 1}), {'code': self.CODE})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(models.Avatar.objects.get(owner__username='test').code, self.CODE)

    def test_update_code(self):
        c = self.login()
        models.Avatar(owner=self.user, code='test', game=self.game).save()
        response = c.post(reverse('aimmo/code', kwargs={'id': 1}), {'code': self.CODE})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(models.Avatar.objects.get(owner__username='test').code, self.CODE)

    def test_code_for_non_existant_game(self):
        c = self.login()
        response = c.post(reverse('aimmo/code', kwargs={'id': 2}), {'code': self.CODE})
        self.assertEqual(response.status_code, 404)

    def test_code_for_non_authed_user(self):
        self.game.public = False
        self.game.save()
        c = self.login()
        response = c.post(reverse('aimmo/code', kwargs={'id': 1}), {'code': self.CODE})
        self.assertEqual(response.status_code, 404)

    def test_default_code(self):
        c = self.login()
        response = c.get(reverse('aimmo/code', kwargs={'id': 1}))
        self.assertEqual(response.status_code, 200)
        self.assertTrue(json.loads(response.content)['code'].startswith('class Avatar'))

    def test_retrieve_code(self):
        models.Avatar(owner=self.user, code=self.CODE, game=self.game).save()
        c = self.login()
        response = c.get(reverse('aimmo/code', kwargs={'id': 1}))
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'code': self.CODE})

    def _associate_game_as_level_num(self, level_num=1, user=None, game=None):
        if game is None:
            game = self.game
        if user is None:
            user = self.user
        models.LevelAttempt(user=user, game=game, level_number=level_num).save()
        models.Game(name='Wrong').save()

    def test_play(self):
        c = self.login()
        response = c.get(reverse('aimmo/play', kwargs={'id': 1}))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['current_user_player_key'], self.user.pk)
        self.assertEqual(response.context['game_url_base'], 'base 1')
        self.assertEqual(response.context['game_url_path'], 'path 1')

    def test_play_for_non_existent_game(self):
        c = self.login()
        response = c.get(reverse('aimmo/play', kwargs={'id': 2}))
        self.assertEqual(response.status_code, 404)

    def _run_test_for_unauthorised_user(self, link, kwarg_name, status_code):
        self._make_game_private()
        c = self.login()
        response = c.get(reverse(link, kwargs={kwarg_name: 1}))
        self.assertEqual(response.status_code, status_code)

    def test_play_for_unauthorised_user(self):
        self._run_test_for_unauthorised_user('aimmo/play', 'id', 404)

    def test_play_inactive_level(self):
        c = self.login()
        self.game.completed = True
        if self.game.is_active:
            self.skipTest('Completed game is active')
        self.game.static_data = '{"test": 1}'
        self.game.save()
        response = c.get(reverse('aimmo/play', kwargs={'id': 1}))
        self.assertEqual(response.status_code, 200)
        self.assertFalse(response.context['active'])
        self.assertEqual(response.context['static_data'], '{"test": 1}')

    def test_games_api(self):
        self.game.main_user = self.user
        self.game.save()
        user2 = User.objects.create_user(username='2', password='password')
        user3 = User.objects.create_user(username='3', password='password')
        models.Avatar(owner=self.user, code=self.CODE, pk=1, game=self.game).save()
        models.Avatar(owner=user2, code='test2', pk=2, game=self.game).save()
        models.Avatar(owner=user3, code='test3', pk=3, game=self.game).save()
        c = Client()
        response = c.get(reverse('aimmo/game_details', kwargs={'id': 1}))
        self.assertJSONEqual(response.content, self.EXPECTED_GAMES)

    def test_games_api_for_non_existent_game(self):
        response = self._go_to_page('aimmo/game_details', 'id', 5)
        self.assertEqual(response.status_code, 404)

    def _run_mark_complete_test(self, request_method, game_id, success_expected):
        c = Client()
        if request_method == "POST":
            response = c.post(reverse('aimmo/complete_game', kwargs={'id': game_id}))
        else:
            response = c.get(reverse('aimmo/complete_game', kwargs={'id': game_id}))
        self.assertEqual(response.status_code == 200, success_expected)
        self.assertEqual(models.Game.objects.get(id=1).completed, success_expected)

    def test_mark_complete(self):
        self._run_mark_complete_test("POST", 1, True)

    def test_mark_complete_for_non_existent_game(self):
        self._run_mark_complete_test("POST", 3, False)

    def test_mark_complete_requires_POST(self):
        self._run_mark_complete_test("GET", 1, False)

    def test_mark_complete_has_no_csrf_check(self):
        c = Client(enforce_csrf_checks=True)
        response = c.post(reverse('aimmo/complete_game', kwargs={'id': 1}))
        self.assertEqual(response.status_code, 200)
        self.assertTrue(models.Game.objects.get(id=1).completed)

    def test_mark_complete_with_data(self):
        c = Client()
        c.post(reverse('aimmo/complete_game', kwargs={'id': 1}), 'static', content_type='application/json')
        self.assertEqual(models.Game.objects.get(id=1).static_data, 'static')

    def test_current_avatar_api_for_non_existent_game(self):
        response = self._go_to_page('aimmo/current_avatar_in_game', 'game_id', 1)
        self.assertEqual(response.status_code, 404)

    def test_current_avatar_api_for_unauthorised_games(self):
        self._run_test_for_unauthorised_user('aimmo/current_avatar_in_game', 'game_id', 401)

    def test_current_avatar_api_for_two_users(self):
        # Set up the first avatar
        first_user = self.user
        models.Avatar(owner=first_user, code=self.CODE, game=self.game).save()
        client_one = self.login()

        # Set up the second avatar
        second_user = User.objects.create_user('test2', 'test2@example.com', 'password2')
        second_user.save()
        models.Avatar(owner=second_user, code=self.CODE, game=self.game).save()
        client_two = Client()
        client_two.login(username='test2', password='password2')

        self.game.public = True
        self.game.can_play = [first_user, second_user]
        self.game.save()

        first_response = client_one.get(reverse('aimmo/current_avatar_in_game', kwargs={'game_id': 1}))
        second_response = client_two.get(reverse('aimmo/current_avatar_in_game', kwargs={'game_id': 1}))

        # Status code starts with 2, success response can be different than 200.
        self.assertEqual(str(first_response.status_code)[0], "2")
        self.assertEqual(str(second_response.status_code)[0], "2")

        # JSON is returned as string so needs to be evaluated.
        first_id = ast.literal_eval(first_response.content)['current_avatar_id']
        second_id = ast.literal_eval(second_response.content)['current_avatar_id']

        self.assertEqual(first_id, 1)
        self.assertEqual(second_id, 2)

    def test_current_avatar_api_call_returns_404_for_logged_out_user(self):
        user = self.user
        models.Avatar(owner=user, code=self.CODE, game=self.game).save()
        client_one = Client()

        self.game.public = True
        self.game.can_play = [user]
        self.game.save()

        first_response = client_one.get(reverse('aimmo/current_avatar_in_game', kwargs={'game_id': 1}))

        self.assertEqual(first_response.status_code, 404)

    def test_id_of_current_avatar_same_as_games_url(self):
        """
        Ensures that the id's are consistent throughout the project. Check for ID's received
        by the current_avatar URL as well as the games URL api.
        """
        user = self.user
        models.Avatar(owner=user, code=self.CODE, game=self.game).save()
        client = self.login()

        self.game.public = True
        self.game.can_play = [user]
        self.game.save()

        current_avatar_api_response = client.get(reverse('aimmo/current_avatar_in_game', kwargs={'game_id': 1}))
        games_api_response = client.get(reverse('aimmo/game_details', kwargs={'id': 1}))

        current_avatar_id = ast.literal_eval(current_avatar_api_response.content)['current_avatar_id']
        games_api_users = json.loads(games_api_response.content)['main']['users']

        self.assertEqual(current_avatar_id, 1)
        self.assertEqual(len(games_api_users), 1)
        self.assertEqual(games_api_users[0]['id'], 1)



from abc import ABCMeta, abstractmethod

import math

INVULNERABILITY_RESISTANCE = 1000


class _Effect(object):
    __metaclass__ = ABCMeta

    def __init__(self, avatar):
        self._avatar = avatar
        self.is_expired = False

    @abstractmethod
    def on_turn(self):
        raise NotImplementedError()


class _TimedEffect(_Effect):
    __metaclass__ = ABCMeta
    EFFECT_TIME = 10

    def __init__(self, *args):
        super(_TimedEffect, self).__init__(*args)
        self._time_remaining = self.EFFECT_TIME

    def remove(self):
        try:
            self._avatar.effects.remove(self)
        except KeyError as e:
            raise KeyError("The avatar object does not exist! Cannot remove the effect.")

    def on_turn(self):
        self._time_remaining -= 1
        if self._time_remaining <= 0:
            self.is_expired = True


class InvulnerabilityPickupEffect(_TimedEffect):
    def __init__(self, *args):
        super(InvulnerabilityPickupEffect, self).__init__(*args)
        self._avatar.resistance += INVULNERABILITY_RESISTANCE

    def remove(self):
        super(InvulnerabilityPickupEffect, self).remove()
        self._avatar.resistance -= INVULNERABILITY_RESISTANCE


class DamageBoostPickupEffect(_TimedEffect):
    def __init__(self, damage_boost, *args):
        assert not math.isinf(damage_boost)

        self._damage_boost = int(round(damage_boost))
        super(DamageBoostPickupEffect, self).__init__(*args)
        self._avatar.attack_strength += self._damage_boost

    def remove(self):
        super(DamageBoostPickupEffect, self).remove()
        self._avatar.attack_strength -= self._damage_boost



from .spawn_location_finder import SpawnLocationFinder
from .map_updaters import ScoreLocationUpdater, MapContext, PickupUpdater, MapExpander

__all__ = [
    'SpawnLocationFinder',
    'ScoreLocationUpdater',
    'MapContext',
    'PickupUpdater',
    'MapExpander'
]



from collections import namedtuple

ReceivedAttackEvent = namedtuple(
    'ReceivedAttackEvent', ['attacking_avatar', 'damage_dealt'])

PerformedAttackEvent = namedtuple(
    'PerformedAttackEvent',
    ['attacked_avatar', 'target_location', 'damage_dealt'])

FailedAttackEvent = namedtuple(
    'FailedAttackEvent', ['target_location'])

MovedEvent = namedtuple(
    'MovedEvent', ['source_location', 'target_location'])

FailedMoveEvent = namedtuple(
    'FailedMoveEvent', ['source_location', 'target_location'])



from unittest import TestCase

from simulation.cell import Cell
from .test_world_map import Serialiser


class TestCell(TestCase):
    def test_equal(self):
        cell1 = Cell(1)
        cell2 = Cell(1)
        self.assertEqual(cell1, cell2)

    def test_not_equal(self):
        cell1 = Cell(1)
        cell2 = Cell(2)
        self.assertNotEqual(cell1, cell2)

    def _create_full_cell(self):
        cell = Cell(Serialiser('location'), False, True)
        cell.avatar = Serialiser('avatar')
        cell.pickup = Serialiser('pickup')
        self.expected = {
            'avatar': 'avatar',
            'generates_score': True,
            'habitable': False,
            'location': 'location',
            'pickup': 'pickup',
            'partially_fogged': False
        }
        return cell

    def test_serialise(self):
        cell = self._create_full_cell()
        self.assertEqual(cell.serialise(), self.expected)

    def test_serialise_no_avatar(self):
        cell = self._create_full_cell()
        cell.avatar = None
        self.expected['avatar'] = None
        self.assertEqual(cell.serialise(), self.expected)

    def test_serialise_no_pickup(self):
        cell = self._create_full_cell()
        cell.pickup = None
        self.expected['pickup'] = None
        self.assertEqual(cell.serialise(), self.expected)



DEFAULT_LEVEL_SETTINGS = {
    'TARGET_NUM_CELLS_PER_AVATAR': 0,
    'TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR': 0,
    'SCORE_DESPAWN_CHANCE': 0,
    'TARGET_NUM_PICKUPS_PER_AVATAR': 0,
    'PICKUP_SPAWN_CHANCE': 0,
    'NO_FOG_OF_WAR_DISTANCE': 1000,
    'PARTIAL_FOG_OF_WAR_DISTANCE': 1000,
}



from simulation.worker_managers.worker_manager import WorkerManager


class ConcreteWorkerManager(WorkerManager):
    def __init__(self, *args, **kwargs):
        self.final_workers = set()
        self.removed_workers = []
        self.updated_workers = []
        self.added_workers = []

        super(ConcreteWorkerManager, self).__init__(*args, **kwargs)

    def create_worker(self, player_id):
        self.added_workers.append(player_id)
        self.final_workers.add(player_id)

    def remove_worker(self, player_id):
        self.removed_workers.append(player_id)
        try:
            self.final_workers.remove(player_id)
        except KeyError:
            pass

    def update_code(self, user):
        self.updated_workers.append(user['id'])
        super(ConcreteWorkerManager, self).update_code(user)



from django.conf.urls import url
from django.contrib.auth import views as auth_views
from django.contrib.auth.decorators import login_required
from django.views.generic import TemplateView
from django.views.generic import RedirectView
from django_js_reverse.views import urls_js

from aimmo import views
from app_settings import preview_user_required

HOMEPAGE_REGEX = r'^aimmo/'

urlpatterns = [
    url(r'^$', login_required(preview_user_required(TemplateView.as_view(template_name='players/home.html'))), name='aimmo/home'),

    url(r'^accounts/login/$', auth_views.login, name='aimmo/login'),
    url(r'^accounts/logout/$', auth_views.logout, {'next_page': 'aimmo/logout_success'}, name='aimmo/logout'),
    url(r'^accounts/logout_success/$', TemplateView.as_view(template_name='registration/success_logout.html'), name='aimmo/logout_success'),

    url(r'^play/(?P<id>[0-9]+)/$', login_required(preview_user_required(views.watch_game)), name='aimmo/play'),
    url(r'^statistics/$', TemplateView.as_view(template_name='players/statistics.html'), name='aimmo/statistics'),

    url(r'^api/csrf_token', views.csrfToken, name='aimmo/csrf_token'),
    url(r'^api/code/(?P<id>[0-9]+)/$', views.code, name='aimmo/code'),
    url(r'^api/games/$', views.list_games, name='aimmo/games'),
    url(r'^api/games/(?P<id>[0-9]+)/$', views.get_game, name='aimmo/game_details'),
    url(r'^api/games/(?P<game_id>[0-9]+)/connection_parameters/$', views.connection_parameters, name='aimmo/connection_parameters'),
    url(r'^api/games/(?P<id>[0-9]+)/complete/$', views.mark_game_complete, name='aimmo/complete_game'),
    url(r'^api/games/(?P<game_id>[0-9]+)/current_avatar/$', views.current_avatar_in_game, name='aimmo/current_avatar_in_game'),

    url(r'^jsreverse/$', urls_js, name='aimmo/js_reverse'),  # TODO: Pull request to make django_js_reverse.urls
    url(r'^games/new/$', views.add_game, name='aimmo/new_game'),

    # TODO: this is a quickfix for redirecting for the Unity resources
    url(r'^watch/(?P<id>[0-9]+)/(?P<resource>.[0-9A-Za-z/.]+)$',
        RedirectView.as_view(url='/static/unity/%(resource)s', permanent=False)),

]



import django
import logging
import os
import sys
import time
from django.conf import settings
from shell_api import log, run_command, run_command_async
import docker_scripts

sys.path.append("/home/travis/build/ocadotechnology/aimmo")

try:
    if os.environ['CI'] == "true":
        ROOT_DIR_LOCATION = os.environ['TRAVIS_BUILD_DIR']
    else:
        ROOT_DIR_LOCATION = os.path.abspath(os.path.dirname((os.path.dirname(__file__))))
except KeyError:
    ROOT_DIR_LOCATION = os.path.abspath(os.path.dirname((os.path.dirname(__file__))))

_MANAGE_PY = os.path.join(ROOT_DIR_LOCATION, 'example_project', 'manage.py')
_SERVICE_PY = os.path.join(ROOT_DIR_LOCATION, 'aimmo-game-creator', 'service.py')
_FRONTEND_BUNDLER_JS = os.path.join(ROOT_DIR_LOCATION, 'game_frontend', 'djangoBundler.js')

PROCESSES = []


def create_superuser_if_missing(username, password):
    from django.contrib.auth.models import User
    try:
        User.objects.get_by_natural_key(username)
    except User.DoesNotExist:
        log('Creating superuser %s with password %s' % (username, password))
        User.objects.create_superuser(username=username, email='admin@admin.com',
                                      password=password)


def run(use_minikube, server_wait=True, capture_output=False, test_env=False, build_target=None):
    logging.basicConfig()
    if test_env:
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "test_settings")
    else:
        sys.path.append(os.path.join(ROOT_DIR_LOCATION, 'example_project'))
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "example_project.settings")

    django.setup()
    run_command(['pip', 'install', '-e', ROOT_DIR_LOCATION], capture_output=capture_output)

    if not test_env:
        run_command(['python', _MANAGE_PY, 'migrate', '--noinput'], capture_output=capture_output)
        run_command(['python', _MANAGE_PY, 'collectstatic', '--noinput'], capture_output=capture_output)

    create_superuser_if_missing(username='admin', password='admin')

    server_args = []
    if use_minikube:
        parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        sys.path.append(os.path.join(parent_dir, 'aimmo_runner'))

        os.chdir(ROOT_DIR_LOCATION)
        run_command(['pip', 'install', '-r', os.path.join(ROOT_DIR_LOCATION, 'minikube_requirements.txt')],
                    capture_output=capture_output)

        # Import minikube here, so we can install the dependencies first
        from aimmo_runner import minikube
        minikube.start(build_target=build_target)

        server_args.append('0.0.0.0:8000')
        os.environ['AIMMO_MODE'] = 'minikube'
    else:
        time.sleep(2)
        os.environ['AIMMO_MODE'] = 'threads'
        docker_scripts.delete_containers()
        if build_target == 'tester':
            run_command(['python', 'all_tests.py'])
        else:
            docker_scripts.build_docker_images(build_target=build_target)
            docker_scripts.start_game_creator()

    os.environ['NODE_ENV'] = 'development' if settings.DEBUG else 'production'
    server = run_command_async(['python', _MANAGE_PY, 'runserver'] + server_args, capture_output=capture_output)
    frontend_bundler = run_command_async(['node', _FRONTEND_BUNDLER_JS], capture_output=capture_output)
    PROCESSES.append(server)
    PROCESSES.append(frontend_bundler)

    if server_wait is True:
        try:
            game.wait()
        except NameError:
            pass

        server.wait()

    return PROCESSES



from unittest import TestCase
from hypothesis import given, assume
from hypothesis import strategies as st
import math
import asyncio

from .mock_world import MockWorld
from simulation.location import Location
from simulation.pickups import DamageBoostPickup, DAMAGE_BOOST_DEFAULT


class TestDamagePickupsAndEffects(TestCase):
    def setUp(self):
        """
        Mock a game for each test individually. MockWorld() will set up a:
        avatar manager, game state, turn manager and a map generator.
        """
        self.game = MockWorld()
        self.game.game_state.add_avatar(1, Location(0, 0))
        _avatar_spawn_cell = self.game.game_state.world_map.get_cell(Location(0, 0))
        self.initial_attack_strength = _avatar_spawn_cell.avatar.attack_strength
        self.cell = self.game.game_state.world_map.get_cell(Location(1, 0))
        self.loop = asyncio.get_event_loop()

    def test_damage_boost_pickup_can_be_picked_up_default(self):
        pickup_created = DamageBoostPickup(self.cell)
        self.cell.pickup = pickup_created

        self.loop.run_until_complete(self.game.simulation_runner.run_single_turn(self.game.avatar_manager.get_player_id_to_serialised_action()))

        self.assertEqual(self.cell.avatar, self.game.avatar_manager.get_avatar(1))
        self.assertEqual(len(self.game.avatar_manager.get_avatar(1).effects), 1)
        damage_boost_effect = self.game.avatar_manager.get_avatar(1).effects.pop()
        self.assertTrue(isinstance(damage_boost_effect, pickup_created.EFFECT))

    @given(st.integers(min_value=1))
    def test_damage_boost_pickup_can_be_picked_up_custom_integer(self, boost_value):
        self.setUp()
        pickup_created = DamageBoostPickup(self.cell, boost_value)
        self.cell.pickup = pickup_created

        self.loop.run_until_complete(self.game.simulation_runner.run_single_turn(self.game.avatar_manager.get_player_id_to_serialised_action()))

        self.assertEqual(self.cell.avatar, self.game.avatar_manager.get_avatar(1))
        self.assertEqual(len(self.game.avatar_manager.get_avatar(1).effects), 1)
        damage_boost_effect = self.game.avatar_manager.get_avatar(1).effects.pop()
        self.assertTrue(isinstance(damage_boost_effect, pickup_created.EFFECT))

    @given(st.floats(min_value=1))
    def test_damage_boost_pickup_can_be_picked_up_custom_floats(self, boost_value):
        assume(not math.isinf(boost_value))
        self.setUp()
        pickup_created = DamageBoostPickup(self.cell, boost_value)
        self.cell.pickup = pickup_created

        self.loop.run_until_complete(self.game.simulation_runner.run_single_turn(self.game.avatar_manager.get_player_id_to_serialised_action()))

        self.assertEqual(self.cell.avatar, self.game.avatar_manager.get_avatar(1))
        self.assertEqual(len(self.game.avatar_manager.get_avatar(1).effects), 1)
        damage_boost_effect = self.game.avatar_manager.get_avatar(1).effects.pop()
        self.assertTrue(isinstance(damage_boost_effect, pickup_created.EFFECT))

    def test_damage_boost_increases_attack_strength_with_default_integer(self):
        """
        Damage boost with no value parameter provided (ie. default).
        """
        pickup_created = DamageBoostPickup(self.cell)
        self.cell.pickup = pickup_created

        self.loop.run_until_complete(self.game.simulation_runner.run_single_turn(self.game.avatar_manager.get_player_id_to_serialised_action()))

        self.assertTrue(self.cell.avatar.attack_strength, self.initial_attack_strength + DAMAGE_BOOST_DEFAULT)

    @given(st.integers(min_value=1))
    def test_damage_boost_increases_attack_strength_with_custom_integers(self, boost_value):
        """
        Damage Boost with random integers provided as a parameter.
        """
        self.setUp()
        pickup_created = DamageBoostPickup(self.cell, boost_value)
        self.cell.pickup = pickup_created

        self.loop.run_until_complete(self.game.simulation_runner.run_single_turn(self.game.avatar_manager.get_player_id_to_serialised_action()))

        self.assertTrue(self.cell.avatar.attack_strength, self.initial_attack_strength + boost_value)



from __future__ import absolute_import

import abc
from unittest import TestCase

from simulation import effects, pickups
from .dummy_avatar import DummyAvatar
from .maps import MockCell


class _BaseCases(object):
    class BasePickupTestCase(TestCase):
        __metaclass__ = abc.ABCMeta

        def setUp(self):
            self.avatar = DummyAvatar(1, None)
            self.cell = MockCell()
            self.pickup = self.pickup_class(self.cell)

        def set_custom_pickup(self, *args):
            self.pickup = self.pickup_class(self.cell, *args)

        def apply_pickup(self):
            self.pickup.apply(self.avatar)

        def apply_second_pickup(self, pickup_two):
            pickup_two.apply(self.avatar)

        def test_pickup_removed(self):
            self.apply_pickup()
            self.assertIs(self.cell.pickup, None)

        @abc.abstractproperty
        def pickup_class(self):
            pass

    class BasePickupEffectTestCase(BasePickupTestCase):
        __metaclass__ = abc.ABCMeta

        @abc.abstractproperty
        def effect_class(self):
            pass

        def test_effect_added(self):
            self.apply_pickup()
            self.assertEqual(len(self.avatar.effects), 1)
            self.assertIsInstance(list(self.avatar.effects)[0], self.effect_class)


class TestHealthPickup(_BaseCases.BasePickupTestCase):
    pickup_class = pickups.HealthPickup

    def test_invulnerability_is_applied(self):
        self.apply_pickup()
        self.assertIsInstance(self.pickup, pickups.HealthPickup)

    def test_health_increases(self):
        self.apply_pickup()
        self.assertEqual(self.avatar.health, 8)

    def test_given_custom_health_pickup_increase_health_increases(self):
        # Test 1 - increase of 5
        self.assertEqual(self.avatar.health, 5)
        self.set_custom_pickup(5)

        self.apply_pickup()

        self.assertEqual(self.avatar.health, 10)

        # Test 2 - increase of 50
        self.avatar.health = 10
        self.set_custom_pickup(50)

        self.apply_pickup()

        self.assertEqual(self.avatar.health, 60)

    def test_health_argument_raises_exception_when_health_exceeded(self):
        self.assertRaises(ValueError, self.set_custom_pickup, 150)

    def test_health_argument_raises_exception_when_health_useless(self):
        self.assertRaises(ValueError, self.set_custom_pickup, 0)

    def test_health_argument_raises_exception_when_health_under(self):
        self.assertRaises(ValueError, self.set_custom_pickup, -50)

    def test_health_cannot_be_greater_than_100(self):
        self.assertEqual(self.avatar.health, 5)
        self.set_custom_pickup(100)

        self.apply_pickup()

        self.assertEqual(self.avatar.health, 100)

    def test_health_rounding_up_to_nearest_int(self):
        self.set_custom_pickup(9.5)
        self.assertEqual(self.pickup.health_restored, 10)

    def test_serialise(self):
        self.assertEqual(self.pickup.serialise(), {
            'type': 'health',
            'location': {
                'x': 0,
                'y': 0,
            }
        })


class TestInvulnerabilityPickup(_BaseCases.BasePickupEffectTestCase):
    pickup_class = pickups.InvulnerabilityPickup
    effect_class = effects.InvulnerabilityPickupEffect

    def test_invulnerability_is_applied(self):
        self.apply_pickup()
        self.assertIsInstance(self.pickup, pickups.InvulnerabilityPickup)

    def test_second_invulnerability_can_be_picked_up(self):
        self.assertEqual(self.avatar.resistance, 0)
        second_pickup = self.pickup_class(MockCell())

        self.apply_pickup()
        self.assertEqual(self.avatar.resistance, 1000)
        self.apply_second_pickup(second_pickup)

        self.assertEqual(self.avatar.resistance, 2000)

    def test_serialise(self):
        self.assertEqual(self.pickup.serialise(), {
            'type': 'invulnerability',
            'location': {
                'x': 0,
                'y': 0,
            }
        })


class TestDamageBoostPickup(_BaseCases.BasePickupEffectTestCase):
    pickup_class = pickups.DamageBoostPickup
    effect_class = effects.DamageBoostPickupEffect

    def test_damage_boost_pickup_is_applied(self):
        self.apply_pickup()
        self.assertIsInstance(self.pickup, pickups.DamageBoostPickup)

    def test_damage_boost_pickup_default_params(self):
        self.assertEqual(len(self.avatar.effects), 0)
        self.apply_pickup()
        self.assertEqual(len(self.avatar.effects), 1)

    def test_serialise(self):
        self.assertEqual(self.pickup.serialise(), {
            'type': 'damage_boost',
            'location': {
                'x': 0,
                'y': 0,
            }
        })



from simulation.action import MoveAction


class Cell(object):
    """
    Any position on the world grid.
    """

    def __init__(self, location, habitable=True, generates_score=False,
                 partially_fogged=False):
        self.location = location
        self.habitable = habitable
        self.generates_score = generates_score
        self.avatar = None
        self.pickup = None
        self.partially_fogged = partially_fogged
        self.actions = []

    def __repr__(self):
        return 'Cell({} h={} s={} a={} p={} f{})'.format(
            self.location, self.habitable, self.generates_score, self.avatar, self.pickup,
            self.partially_fogged)

    def __eq__(self, other):
        return self.location == other.location

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.location)

    @property
    def moves(self):
        return [move for move in self.actions if isinstance(move, MoveAction)]

    @property
    def is_occupied(self):
        return self.avatar is not None

    def serialise(self):
        if self.partially_fogged:
            return {
                'generates_score': self.generates_score,
                'location': self.location.serialise(),
                'partially_fogged': self.partially_fogged
            }
        else:
            return {
                'avatar': self.avatar.serialise() if self.avatar else None,
                'generates_score': self.generates_score,
                'habitable': self.habitable,
                'location': self.location.serialise(),
                'pickup': self.pickup.serialise() if self.pickup else None,
                'partially_fogged': self.partially_fogged
            }



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup


setup(
    name='aimmo_runner',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'django >= 1.8.3, < 1.9.0',
        'django-autoconfig >= 0.3.6, < 1.0.0',
        'django-forms-bootstrap',
        'django-js-reverse',
        'eventlet',
        'flask',
        'flask-socketio',
        'requests',
        'six',
        'hypothesis',
        'flask-cors',
        'psutil',
        'docker >= 3.5, < 3.6',
        'kubernetes == 5.0.0',
    ],
    zip_safe=False,
)



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup
import sys
import os


withcoverage = os.environ.get('WITH_COVERAGE')

setup(
    name='aimmo-game',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'eventlet',
        'flask',
        'flask-cors',
        'python-socketio==2.0.0',
        'requests',
        'six',
        'kubernetes'
    ],
    setup_requires=[
        "pytest-runner"
    ],
    tests_require=[
        'pytest==4.0.2',
        'pytest-asyncio',
        'asynctest',
        'httmock',
        'mock',
        'hypothesis',
        'docker',
        'aiohttp',
        'aiohttp_cors'
    ],
    test_suite='tests',
    zip_safe=False,
)



"""
Any helper functions used for the unity game.
"""

from django.shortcuts import render, get_object_or_404
from django.contrib.auth.models import User
from aimmo import app_settings, exceptions
from models import Game


def render_game(request, game):
    """
    :param request: Request object used to generate this response.
    :param game: Game object.
    :return: HttpResponse object with a given context dictionary and a template.
    """
    context = {'current_user_player_key': request.user.pk, 'active': game.is_active,
               'static_data': game.static_data or '{}'
               }

    connection_settings = get_environment_connection_settings(game.id)

    context.update(connection_settings)

    return render(request, 'players/game_ide.html', context)


def get_environment_connection_settings(game_id):
    """
    This function will return the correct URL parts and a SSL flag
    based on the environment of which it exists in.

    :param request: Request object used to generate this response.
    :param game_id: Integer with the ID of the game.
    :return: A dict object with all relevant settings.
    """

    return {
        'game_url_base': _add_game_port_to_game_base(game_id),
        'game_url_path': app_settings.GAME_SERVER_URL_FUNCTION(game_id)[1],
        'game_ssl_flag': app_settings.GAME_SERVER_SSL_FLAG, 'game_id': game_id
    }


def _add_game_port_to_game_base(game_id):
    game_base = app_settings.GAME_SERVER_URL_FUNCTION(game_id)[0]
    game_port = app_settings.GAME_SERVER_PORT_FUNCTION(game_id)

    if _connection_on_k8s_mode(game_port):
        return game_base

    return "{0}:{1}".format(game_base, game_port)


def _connection_on_k8s_mode(game_port):
    return game_port == 0


def get_avatar_id_from_user(user, game_id):
    """
    A helper function which will return an avatar ID that is assigned to a
    particular django user owner in a game.
    :param user: A django user object taken from the request.
    :param game_id: The game ID in which the avatar is being requested from.
    :return: An integer containing the avatar_ID.
    """
    game = get_object_or_404(Game, id=game_id)
    if not game.can_user_play(user):
        raise exceptions.UserCannotPlayGameException

    avatar = game.avatar_set.get(owner=user.id)

    return avatar.id
